#include "ColormapRenderer.h"
#include "ScalarAttributes.h"
#include "Grid.h"
#include <GLUT/glut.h>									//GLUT library
#include <algorithm>
#include <math.h>


using namespace std;



void rainbow(float value,float& R,float& G,float& B)	//Maps a scalar in [0,1] to a color using the rainbow colormap
{
   const float dx=0.8f;

   value = (6-2*dx)*value+dx;
   R = max(0.0f,(3-(float)fabs(value-4)-(float)fabs(value-5))/2);
   G = max(0.0f,(4-(float)fabs(value-2)-(float)fabs(value-4))/2);
   B = max(0.0f,(3-(float)fabs(value-1)-(float)fabs(value-2))/2);
}

void grayscale(float value,float& R,float& G,float& B)	//Maps a scalar in [0,1] to a color using a grayscale colormap
{
   R = G = B = value;
}

void saturation(float value,float& R,float& G,float& B)	//Maps a scalar in [0,1] to a color using a saturation colormap
{
   float color_r = 0;									//The base color whose saturation we change (green).	
   float color_g = 1;									//Try different colors!	
   float color_b = 0;
   
   if (value<0.5)										//value in [0,0.5]: modulate the luminance from black to the base-color.
   {   
	   R = 2*value*color_r;
	   G = 2*value*color_g;
	   B = 2*value*color_b;
   }
   else													//value in [0.5,1]: modulate the saturation from base-color to white.
   {
	   value = 2*(value-0.5);	
	   R = (1-value)*color_r + 1*value;
	   G = (1-value)*color_g + 1*value;
	   B = (1-value)*color_b + 1*value;
   }
}


void ColormapRenderer::mapToColor(float value,float& R,float& G,float& B)
{														//Map v to a color using currently active colormap
	switch (colormap_type)
	{
	case COLORMAP_GRAYSCALE: 
		grayscale(value,R,G,B);
		break;
	case COLORMAP_RAINBOW: 
		rainbow(value,R,G,B);
		break;
	case COLORMAP_SATURATION: 
		saturation(value,R,G,B);
		break;
	}
}		

void ColormapRenderer::drawC0Cells(Grid& g)				//Draw 'g' by showing C0-interpolated scalar values over cells.
{														//
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);			//1.    Render cells filled
	glShadeModel(GL_FLAT);								//2.    Do not interpolate vertex colors (use flat-shaded polygons)

	ScalarAttributes& scalars = g.cellScalars();		//3.    Get scalar range, to normalize scalar-values next in [0,1]
	float scalar_min = scalars.getMin();
	float scalar_max = scalars.getMax();

	for(int i=0;i<g.numCells();++i)						//4. Draw all cells in the grid as filled flat-shaded polygons
	{
		int   cell[10];									//Should have enough storage space for our largest cell
		
		int nvertices = g.getCell(i,cell);				//4.1. Get the 'nvertices' vertex-IDs of i-th cell
		if (nvertices!=3 && nvertices!=4)				//     We only handle here drawing of triangle and quad cells.
		   continue;									//     It is quite simple to extend this to other cell types.

		if (nvertices==3)								//     Triangle cells:
			glBegin(GL_TRIANGLES);						
		else //nvertices==4	
			glBegin(GL_QUADS);							//     Quad cells:
				
		float value = scalars.getC0Scalar(i);			//4.2. Get cell scalar value, and normalize it in [0,1]														
		value = (value-scalar_min)/(scalar_max-scalar_min);	
		
		float R,G,B;									//4.3. Map scalar value to color
		mapToColor(value,R,G,B);						//4.4. Pass color to OpenGL. As we use flat-shading, the entire cell
		glColor3f(R,G,B);								//     will have this color	

		for(int j=0;j<nvertices;++j)					//4.5. Render current cell
		{
			float x[2];
			g.getPoint(cell[j],x);						//     Get vertex coordinates of j-th vertex of i-th cell				
			glVertex3f(x[0],x[1],0);					//     Pass this coordinate to OpenGL
		}	
		glEnd();                                        //4.6. Finish rendering current cell
	}
}



void ColormapRenderer::draw(Grid& g)					//Draw grid 'g' depending on drawing style attribute-value
{
	glClearColor(1,1,1,1);								//1. Clear the frame and depth buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	drawC0Cells(g);
	
	glutSwapBuffers();
}




