#include <GLUT/glut.h>
#include <math.h>
#include <iostream>
#include <list>
#include "UniformGrid.h"
#include "TensorVolume.h"
#include "ColormapRenderer.h"
#include "DirCodingRenderer.h"
#include "pca.h"
#include "Point3d.h"

using namespace std;




TensorVolume<float>  tensor;							//Tensor field to visuaize
Volume<float>		 fa;								//Fractional anisotropy (Basser and Pierpaoli '96)
Volume<float>		 diffusivity;						//Mean diffusivity
Volume<Point3d>		 e0;								//Major eigenvector
list<pair<string,Volume<float>*> > metrics;				//The metrics we can visualize
list<pair<string,Volume<float>*> >::iterator 
					 current_metric;					//The current metric we visualize
bool				 confidence_mask = true;			//Whether to mask the displayed metrics with the confidence field or not

	
UniformGrid*		 slice			 = 0;				//Slice through the various tensor metrics volumes
ColormapRenderer	 col_renderer;						//Simple colormap renderer for the slice's scalars
DirCodingRenderer	 dir_renderer;						//Simple directional-color-coding renderer for the slice's vectors
Renderer*			 renderer		 = 0;				//Active renderer (either colormap or directional color-coding)
float				 l0_max;							//Maximum value of the major eigenvalue (used for major-eigenvector normalization)

int					 colormap	     = 0;				//Colormap to use when rendering the slice	
int					 slice_direction;					//Slicing direction: 0=XY, 1=YZ, 2=XZ
int					 slice_position;					//Slice position along the current slicing direction
int					 slice_range;						//Maximum allowed slice_position value for current slicing direction
const char*			 input_file		 = 0;				//Name of input NRRD tensor file





inline void setSliceData(UniformGrid& slice,Volume<float>& scalars,Volume<Point3d>& vectors,int i,int j,int k,int offset)
{														//Copy data from a 3D scalar and a 3D vector volume onto a 2D slice
	float   s = scalars(i,j,k);	
	Point3d v = vectors(i,j,k);
	
	if (confidence_mask)								//Mask the values with the confidence field, if so desired
	{													
	  float conf = tensor.confidence(i,j,k);
	  s *= conf;										//Mask scalar data
	  v *= pow(conf/l0_max,0.7f);						//Mask (and normalize) vector data. The power function achieves a nice
														//correction of the luminance values, needed to see both small and large vectors
	}
	slice.cellScalars().setC0Scalar(offset,s);			//Copy the scalar data from the volume to the slice
	slice.cellVectors().setC0Vector(offset,v.data);		//Copy the vector data from the volume to the slice
}




void extractSlice(Volume<float>& data, UniformGrid*& slice, int slice_direction, int slice_position)
{														//Extract a given 2D slice from a 3D scalar and 3D vector volume

	delete slice;										//1. Delete previous since since we make a new one

	int offset=0;
	switch(slice_direction)
	{
	case 0:												//Slice data in XY plane (orthogonal to Z axis)
	   {
		  int size1 = data.getWidth(), size2 = data.getHeight();
		  float sM = max(size1,size2);
	      slice = new UniformGrid(size1+1,size2+1,0,0,size1/sM,size2/sM);
		  for(int j=0;j<size2;++j)
		  for(int i=0;i<size1;++i)
		     setSliceData(*slice,data,e0,i,j,slice_position,offset++); 
		  break;	 
	   }
	
	case 1:												//Slice data in YZ plane (orthogonal to X axis)
	   {
		  int size1 = data.getHeight(), size2 = data.getDepth();	   
		  float sM = max(size1,size2);
	      slice = new UniformGrid(size1+1,size2+1,0,0,size1/sM,size2/sM);
		  for(int j=0;j<size2;++j)
		  for(int i=0;i<size1;++i)
		     setSliceData(*slice,data,e0,slice_position,i,size2-j-1,offset++); 
		  break;	 
	   }
		
	case 2:												//Slice data in XZ plane (orthogonal to Y axis)
	   {
		  int size1 = data.getWidth(), size2 = data.getDepth();
		  float sM = max(size1,size2);
	      slice = new UniformGrid(size1+1,size2+1,0,0,size1/sM,size2/sM);
		  for(int j=0;j<size2;++j)
		  for(int i=0;i<size1;++i)
		     setSliceData(*slice,data,e0,i,slice_position,size2-j-1,offset++); 
		  break;	 
	   }
	}
}


void saveMetrics()										//Save the various computed metrics as VTK files. 
{														//Modify this as needed to e.g. save other metrics or save in other formats.	
	const char* fa_file   = "fractional_anisotropy.vtk";
	const char* diff_file = "mean_diffusivity.vtk";
	const char* e0_file   = "major_eigenvector.vtk";

	cout<<"Saving tensor metrics:"<<endl<<endl;
	cout<<" -fractional anisotropy: "<<fa_file<<endl;
	cout<<" -mean diffusivity:      "<<diff_file<<endl;
	cout<<" -major eigenvector:     "<<e0_file<<endl;
		
	fa.exportVTKField(fa_file);
	diffusivity.exportVTKField(diff_file);
	e0.exportVTKField(e0_file);
}



void viewing(int W, int H)								//Window resize function, sets up viewing parameters (GLUT callback function)
{
	glMatrixMode(GL_MODELVIEW);							//1. Set the modelview matrix (including the camera position and view direction)
	glLoadIdentity();										

	glMatrixMode(GL_PROJECTION);						//2. Set the projection matrix
	glLoadIdentity();
	glOrtho(0,1,0,1,-1,1);

	glViewport(0,0,W,H);								//3. Set the viewport to the entire size of the rendering window
}



void draw()												//Render the grid (GLUT callback function)
{
	renderer->draw(*slice);
}


void keyboard(unsigned char c,int,int)					//Press space to cycle through the renderer's drawing styles and colormaps
{
    bool need_slicing = false;

	switch(c)
	{
	case ' ':											//Change slicing direction along the three axes
		{
		  slice_direction = (slice_direction+1) % 3;
		  const char* slice_name;		  
		  switch (slice_direction)
		  {
		  case 0:										//Direction 0: slice XY, along Z
			slice_range = tensor.getDepth();
			slice_name  = "XY";
			break;
		  case 1:										//Direction 1: slice YZ, along X
			slice_range = tensor.getWidth();
			slice_name  = "YZ";
			break;
		  case 2:										//Direction 2: slice ZX, along Y	
			slice_range = tensor.getHeight();
			slice_name  = "XZ";
			break;
		  }		
		  slice_position  = slice_range/2;
		  cout<<"Slicing direction: "<<slice_name<<", "<<slice_range<<" slices"<<endl;
		  need_slicing    = true;
		  break;	
		}
	case 'C':											//Change colormap used to visualize the slice's scalars
		{
		  if (current_metric!=metrics.end())			//Only do this if we are not visualizing the direction color-coded eigenvectors
		  {
		    colormap = (++colormap) % 3;
		    col_renderer.setColormapType((ColormapRenderer::COLORMAP_TYPE)colormap);
		  }	
		  break;
		}
	case '+':											//Move the slice forward along the slicing direction
	case '=':
		{
		  slice_position = (slice_position+1) % slice_range;
		  need_slicing   = true;
		  break;
		}
	case '-':											//Move the slice backward along the slicing direction
	case '_':
	    {
		  --slice_position;
		  if (slice_position<0) slice_position = slice_range-1;
		  need_slicing = true;
		  break;
		}
	case 'm':											//Change the metric to visualize.
		{
		  if (current_metric==metrics.end())			//
		     current_metric = metrics.begin();
		  else 	 
		     ++current_metric;
		  
		  if (current_metric==metrics.end())			//If 'current_metric' is at the end of metrics[], visualize the direction color-coded
		  {												//eigenvector. For this, we need to change the renderer.
		    cout<<"Showing metric: major-eigenvector direction"<<endl;
			renderer = &dir_renderer;
		  }
		  else 
		  {
			cout<<"Showing metric: "<<current_metric->first<<endl;
			renderer = &col_renderer;
		  }   
		  
		  need_slicing   = true;
		  break;
		}
	case 'c':											//Toggle on/off the masking of the visualized values with the confidence values
		{
		  confidence_mask = !confidence_mask;
		  cout<<"Masking with confidence: "<<((confidence_mask)?"true":"false")<<endl;
		  need_slicing = true;
		  break;
		}  		
	case 's':											//Save the computed metrics to files
		{
		  saveMetrics();
		  break;
		}  	
	}
	
	if (need_slicing)									//If we need to recompute the 2D slice, do it now
	{
	  Volume<float>* data = (current_metric==metrics.end())? metrics.begin()->second : current_metric->second;
      extractSlice(*data,slice,slice_direction,slice_position);
	}   
			
	glutPostRedisplay();    
}	




void computePCA(TensorVolume<float>& f, int i, int j, int k, Point3d& e0, Point3d& e1, Point3d& e2, float& l0, float& l1, float& l2)
{																	//Given a 3D tensor volume, compute its eigenvectors and eigenvalues
																	//at each voxel.
    float *M[3],*V[3],W[3],MM0[3],MM1[3],MM2[3],VV0[3],VV1[3],VV2[3]; 
  
    M[0] = MM0; M[1] = MM1; M[2] = MM2; V[0] = VV0; V[1] = VV1; V[2] = VV2; 

																	//Fill-in symmetric M[][] for eigen-computation
    M[0][0] = f.dxx(i,j,k);  M[0][1] = f.dxy(i,j,k);  M[0][2] = f.dxz(i,j,k);	
    M[1][0] = M[0][1];       M[1][1] = f.dyy(i,j,k);  M[1][2] = f.dyz(i,j,k);
    M[2][0] = M[0][2];		 M[2][1] = M[1][2];		  M[2][2] = f.dzz(i,j,k);
    
    JacobiN(M,3,W,V);												//Compute the eigenvalues W[] and vectors V[] of M[][]
																	//Results are returned in decreasing eigenval-order W[0]>W[1]>W[2]
    W[0] = fabs(W[0]); W[1] = fabs(W[1]); W[2] = fabs(W[2]);
																	//Return eigenvector and eigenvalue results
	e0   = Point3d(VV0[0],VV1[0],VV2[0]); e0.normalize(); e0 *= W[0];
	e1   = Point3d(VV0[1],VV1[1],VV2[1]); e1.normalize(); e1 *= W[1];
	e2   = Point3d(VV0[2],VV1[2],VV2[2]); e2.normalize(); e2 *= W[2];
		
	l0 = W[0];   
	l1 = W[1];	
	l2 = W[2];	
}






void computeMetrics(TensorVolume<float>& tensor,Volume<float>& fa,Volume<float>& diffusivity,Volume<Point3d>& e0_vol)
{														//Compute various tensor-field metrics, and save them for later use
	Point3d e0,e1,e2; float l0,l1,l2;
	
	int D=tensor.getDepth(),H=tensor.getHeight(),W=tensor.getWidth(); 
	
	fa.makeVolume(W,H,D);								//Allocate volumes for the tensor metrics we will compute
	diffusivity.makeVolume(W,H,D);
	e0_vol.makeVolume(W,H,D);

	l0_max = 0;
													
	for(int k=0;k<D;++k)	
	for(int j=0;j<H;++j)	
	for(int i=0;i<W;++i)	
	{
	   computePCA(tensor,i,j,k,e0,e1,e2,l0,l1,l2);
	   
	   float md = (l0+l1+l2)/3;							
	   float FA = sqrt(((l0-md)*(l0-md)+(l1-md)*(l1-md)+(l2-md)*(l2-md))/(2*l0*l0+l1*l1+l2*l2));
															
	   fa(i,j,k)		  = FA;							//Fractional anisotropy metric
	   diffusivity(i,j,k) = md;							//Mean diffusivity metric
	   e0_vol(i,j,k)	  = e0;							//Major eigenvector metric
	   
	   l0_max = max(l0_max,l0);							//Retrieve range of major eigenvector (for later normalization of this vector field)
	}	
}




int main(int argc,char* argv[])							//Main program
{
	cout<<"4. Tensor metrics"<<endl;
	cout<<"In the graphics window, press"<<endl;
	cout<<"   space:   slice the 3D metrics volumes along the XY, YZ, and XZ planes"<<endl;
	cout<<"   +,=:     move slice forward"<<endl;
	cout<<"   -,_:     move slice backward"<<endl;
	cout<<"   C:       change the colormap used for scalar values"<<endl;
	cout<<"   m:       change the shown metric"<<endl;
	cout<<"   c:       toggle masking the visualized data with the measurement confidence"<<endl;
	cout<<"   s:       save the computed metrics to VTK files"<<endl;  


    glutInit(&argc, argv);								//1. Initialize the GLUT toolkit
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);		//2. Ask GLUT to create next windows with a RGB framebuffer and a Z-buffer too
    glutInitWindowSize(500,500);						//3. Tell GLUT how large are the windows we want to create next
    glutCreateWindow("4. Tensor metrics");
	
	

    input_file = (argc>1)? argv[1] : "DATA/gk2-rcc-mask.nhdr";
	
    bool ok = tensor.readNRRD(input_file);				//4. Read the tensor field from an NRRD-format file														
	if (!ok)
	{
		cout<<"Error reading tensor volume "<<input_file<<", aborting"<<endl;
		exit(1);
	}	

	computeMetrics(tensor,fa,diffusivity,e0);			//5. Compute tensor metrics
	
	slice_direction = 0;								//6. Initialize visualization parameters
	slice_position  = tensor.getDepth()/2;
	slice_range     = tensor.getDepth();

														//7. Collect all metrics we will offer for visualization
	metrics.push_back(make_pair("diffusivity",&diffusivity));
	metrics.push_back(make_pair("fractional anisotropy",&fa));
	metrics.push_back(make_pair("confidence",&tensor.confidence));
	current_metric = metrics.begin();
														//8. Initialize the slice to be visualized
	extractSlice(*current_metric->second,slice,slice_direction,slice_position);
	
														//9. Initialize the renderer
	col_renderer.setColormapType((ColormapRenderer::COLORMAP_TYPE)colormap);
	renderer = &col_renderer;
	
	glutKeyboardFunc(keyboard);   
    glutDisplayFunc(draw);								//10.Add a drawing callback to the window	
    glutReshapeFunc(viewing);							//11.Add a resize callback to the window
    glutMainLoop();										//12.Start the event loop that displays the graph and handles window-resize events

	return 0;
}




